import pytest

from aiothornode.env import CHAOS_NET_BNB_ENVIRONMENT, TEST_NET_ENVIRONMENT_MULTI_1


def test_env_copy():
    env = CHAOS_NET_BNB_ENVIRONMENT.copy()
    env.seed_url = 'lol'
    assert CHAOS_NET_BNB_ENVIRONMENT.seed_url is not 'lol'

    with pytest.raises(AssertionError):
        env.set_consensus(0, 0)

    with pytest.raises(AssertionError):
        env.set_consensus(10, 5)

    env.set_consensus(1, 5)
    assert env.consensus_total == 5
    assert env.consensus_min == 1



import asyncio
import logging

import ujson
from aiohttp import ClientSession, ClientConnectorError

THORNODE_PORT = 1317
TENDERMINT_RPC_PORT_TESTNET = 26657
TENDERMINT_RPC_PORT_MAINNET = 27147


class ThorNodeClient:
    @staticmethod
    def connection_url(ip_address, path, port):
        path = '/' + path.lstrip('/')
        return f'http://{ip_address}:{port}{path}'

    def __init__(self, node_ip: str, session: ClientSession, timeout=3.0, logger=None,
                 thornode_port=THORNODE_PORT,
                 tendermint_rpc_port=TENDERMINT_RPC_PORT_MAINNET):
        self.session = session
        self.timeout = timeout
        self.node_ip = node_ip
        self.logger = logger or logging.getLogger('ThorNodeClient')
        self.tendermint_rpc_port = tendermint_rpc_port
        self.thornode_port = thornode_port

    async def request(self, path, is_rpc=False):
        port = self.tendermint_rpc_port if is_rpc else self.thornode_port
        url = self.connection_url(self.node_ip, path, port)
        try:
            async with self.session.get(url, timeout=self.timeout) as resp:
                text = await resp.text()
                return ujson.loads(text)
        except (ClientConnectorError, asyncio.TimeoutError) as e:
            self.logger.warning(f'Cannot connect to THORNode ({self.node_ip}) for "{path}" (err: {e}).')
            return None

    def __repr__(self) -> str:
        return f'ThorNodeClient({self.node_ip!r})'



import aiohttp
import pytest

from aiothornode.connector import ThorConnector
from aiothornode.env import CHAOS_NET_BNB_ENVIRONMENT, TEST_NET_ENVIRONMENT_MULTI_1


@pytest.fixture()
async def session():
    session = aiohttp.ClientSession()
    yield session
    await session.close()


@pytest.fixture()
async def testnet_connector(session):
    con = ThorConnector(TEST_NET_ENVIRONMENT_MULTI_1, session)
    await con.update_nodes()
    return con


@pytest.fixture()
async def chaosnet_connector(session):
    con = ThorConnector(CHAOS_NET_BNB_ENVIRONMENT, session)
    await con.update_nodes()
    return con



from copy import copy
from dataclasses import dataclass, field


class ThorException(Exception):
    def __init__(self, j, *args) -> None:
        super().__init__(*args)
        if j and isinstance(j, dict):
            self.code = j.get('code', '')
            self.message = j.get('message', '')
            self.details = j.get('details', [])
        else:
            self.code = -10000
            self.message = 'Failed to decode the error'
            self.details = [
                repr(j)
            ]


@dataclass
class ThorQueue:
    outbound: str = '0'
    swap: str = '0'
    internal: str = '0'

    @classmethod
    def from_json(cls, j):
        return cls(**j)

    @property
    def total(self):
        return int(self.outbound) + int(self.swap) + int(self.internal)


@dataclass
class ThorNodeAccount:
    STATUS_STANDBY = 'standby'
    STATUS_ACTIVE = 'active'
    STATUS_READY = 'ready'
    STATUS_WHITELISTED = 'whitelisted'
    STATUS_UNKNOWN = 'unknown'
    STATUS_DISABLED = 'disabled'

    node_address: str = 'thor?'
    status: str = ''
    pub_key_set: dict = field(default_factory=dict)
    validator_cons_pub_key: str = ''
    bond: str = ''
    active_block_height: int = 0
    bond_address: str = ''
    status_since: str = ''
    signer_membership: list = field(default_factory=list)
    requested_to_leave: bool = False
    forced_to_leave: bool = False
    leave_height: int = 0
    ip_address: str = ''
    version: str = ''
    slash_points: int = 0
    jail: dict = field(default_factory=dict)
    current_award: str = ''
    observe_chains: list = field(default_factory=list)
    preflight_status: dict = field(default_factory=dict)

    @classmethod
    def from_json(cls, j):
        return cls(**j)

    @property
    def preflight_status_reason_and_code(self):
        status = self.preflight_status.get('status', '').lower()
        reason = self.preflight_status.get('reason', '')
        code = self.preflight_status.get('code', 0)
        return status, reason, code

    @property
    def is_good(self):
        status = self.status.lower()
        return status in (self.STATUS_ACTIVE, self.STATUS_WHITELISTED) \
               and not self.requested_to_leave and not self.forced_to_leave and self.ip_address


@dataclass
class ThorLastBlock:
    chain: str = ''
    last_observed_in: int = 0
    last_signed_out: int = 0
    thorchain: int = 0

    @classmethod
    def from_json(cls, j):
        x = cls()
        x.chain = j.get('chain', '')
        x.thorchain = j.get('thorchain', 0)
        x.last_observed_in = j.get('last_observed_in', 0) if 'last_observed_in' in j else j.get('lastobservedin')
        x.last_signed_out = j.get('last_signed_out', 0) if 'last_signed_out' in j else j.get('lastsignedout')
        return x


@dataclass
class ThorPool:
    STATUS_AVAILABLE = 'Available'
    STATUS_BOOTSTRAP = 'Bootstrap'
    STATUS_ENABLED = 'Enabled'

    balance_asset: str = '0'
    balance_rune: str = '0'
    asset: str = ''
    pool_units: str = '0'
    status: str = ''
    synth_units: str = ''
    decimals: str = '0'
    error: str = ''
    pending_inbound_rune: str = '0'
    pending_inbound_asset: str = '0'

    @property
    def balance_asset_int(self):
        return int(self.balance_asset)

    @property
    def balance_rune_int(self):
        return int(self.balance_rune)

    @property
    def pool_units_int(self):
        return int(self.pool_units)

    @property
    def assets_per_rune(self):
        return self.balance_asset_int / self.balance_rune_int

    @property
    def runes_per_asset(self):
        return self.balance_rune_int / self.balance_asset_int

    @classmethod
    def from_json(cls, j):
        return cls(**j)


@dataclass
class ThorConstants:
    constants: dict = field(default_factory=dict)
    data_types: dict = field(default_factory=dict)

    DATA_TYPES = ('int_64_values', 'bool_values', 'string_values')

    @classmethod
    def from_json(cls, j):
        holder = cls()
        for dt in cls.DATA_TYPES:
            subset = j.get(dt, {})
            holder.data_types[dt] = {}
            for k, v in subset.items():
                holder.constants[k] = v
                holder.data_types[dt][k] = v

        return holder

    def get(self, name, default=None):
        return self.constants.get(name, default)

    def __getitem__(self, item):
        return self.constants[item]


@dataclass
class ThorMimir:
    constants: dict = field(default_factory=dict)

    @classmethod
    def from_json(cls, j: dict):
        holder = cls()
        for k, v in j.items():
            holder.constants[k] = v
        return holder

    def get(self, name, default=None):
        return self.constants.get(name, default)

    def __getitem__(self, item):
        return self.constants[item]


@dataclass
class ThorEnvironment:
    seed_url: str = ''
    midgard_url: str = ''

    timeout: float = 3.0
    consensus_min: int = 2
    consensus_total: int = 3

    path_queue: str = '/thorchain/queue'
    path_nodes: str = '/thorchain/nodes'
    path_pools: str = "/thorchain/pools"
    path_pools_height: str = "/thorchain/pools?height={height}"
    path_pool: str = "/thorchain/pool/{pool}"
    path_pool_height: str = "/thorchain/pool/{pool}?height={height}"

    path_last_blocks: str = "/thorchain/lastblock"
    path_constants: str = "/thorchain/constants"
    path_mimir: str = "/thorchain/mimir"

    def copy(self):
        return copy(self)

    def set_consensus(self, minimum, total):
        assert total >= 1
        assert 1 <= minimum <= total
        self.consensus_total = total
        self.consensus_min = minimum



from aiothornode.types import ThorEnvironment

CHAOS_NET_BNB_ENVIRONMENT = ThorEnvironment(seed_url='https://chaosnet-seed.thorchain.info/',
                                            midgard_url='https://chaosnet-midgard.bepswap.com/',
                                            path_nodes='/thorchain/nodeaccounts')

TEST_NET_ENVIRONMENT_MULTI_1 = ThorEnvironment(seed_url='https://testnet.seed.thorchain.info',
                                               midgard_url='https://testnet.midgard.thorchain.info/')

MULTICHAIN_CHAOSNET_ENVIRONMENT = ThorEnvironment(seed_url='https://seed.thorchain.info/',
                                                  midgard_url='https://midgard.thorchain.info/')



import aiohttp

from aiothornode.connector import *
from aiothornode.env import TEST_NET_ENVIRONMENT_MULTI_1


async def main():
    env = TEST_NET_ENVIRONMENT_MULTI_1

    async with aiohttp.ClientSession() as session:
        connector = ThorConnector(env, session)

        pools = await connector.query_pools(377)
        print(pools[0])
        print(f'Total {len(pools)} pools')


if __name__ == '__main__':
    asyncio.run(main())



import pytest

from aiothornode.types import ThorPool
from .fixtures import *


async def internal_constants_test(c: ThorConnector):
    constants = await c.query_constants()
    assert constants['BlocksPerYear'] > 100
    assert constants.get('blabla', 'lol') == 'lol'
    str_const = constants.get('DefaultPoolStatus')
    assert isinstance(str_const, str) and len(str_const) > 1
    assert constants['BadValidatorRate'] > 10


@pytest.mark.asyncio
async def test_constants(testnet_connector: ThorConnector):
    await internal_constants_test(testnet_connector)


@pytest.mark.asyncio
async def test_constants_chaos(chaosnet_connector: ThorConnector):
    await internal_constants_test(chaosnet_connector)


async def internal_queue_test(c: ThorConnector):
    queue = await c.query_queue()
    assert int(queue.swap) >= 0
    assert int(queue.outbound) >= 0


@pytest.mark.asyncio
async def test_update_nodes(testnet_connector: ThorConnector):
    print(testnet_connector._clients)


@pytest.mark.asyncio
async def test_queue_test(testnet_connector: ThorConnector):
    await internal_queue_test(testnet_connector)


@pytest.mark.asyncio
async def test_queue_chaos(chaosnet_connector: ThorConnector):
    await internal_queue_test(chaosnet_connector)


def pool_validate(pool: ThorPool):
    assert pool
    assert pool.asset
    assert pool.runes_per_asset > 0
    assert pool.assets_per_rune > 0
    assert pool.pool_units_int > 0
    assert pool.balance_asset_int > 0
    assert pool.balance_rune_int > 0
    assert pool.status in (pool.STATUS_AVAILABLE, pool.STATUS_BOOTSTRAP, pool.STATUS_ENABLED)


async def internal_pools_test(c: ThorConnector, height=None):
    pools = await c.query_pools(height)
    # print(pools)
    assert pools, "no pools"
    for pool in pools:
        pool_validate(pool)


async def internal_pool_test(c: ThorConnector, pool, height=None):
    pool = await c.query_pool(pool, height)
    # print(pools)
    assert pool, "no pool"
    pool_validate(pool)


@pytest.mark.asyncio
async def test_pools_test(testnet_connector: ThorConnector):
    await internal_pools_test(testnet_connector)
    await internal_pools_test(testnet_connector, height=36001)
    await internal_pool_test(testnet_connector, 'BTC.BTC')
    await internal_pool_test(testnet_connector, 'BTC.BTC', 36002)


@pytest.mark.asyncio
async def test_pools_chaos(chaosnet_connector: ThorConnector):
    await internal_pools_test(chaosnet_connector)
    await internal_pools_test(chaosnet_connector, height=1000001)
    await internal_pool_test(chaosnet_connector, 'BNB.BNB')
    await internal_pool_test(chaosnet_connector, 'BNB.BNB', height=100001)


async def internal_last_blocks_test(c: ThorConnector):
    lb = await c.query_last_blocks()
    assert lb


@pytest.mark.asyncio
async def test_last_block_test(testnet_connector: ThorConnector):
    await internal_last_blocks_test(testnet_connector)


@pytest.mark.asyncio
async def test_last_block_chaos(chaosnet_connector: ThorConnector):
    await internal_last_blocks_test(chaosnet_connector)



from logging import INFO

import aiohttp

from aiothornode.connector import *
from aiothornode.env import CHAOS_NET_BNB_ENVIRONMENT


async def main():
    env = CHAOS_NET_BNB_ENVIRONMENT

    async with aiohttp.ClientSession() as session:
        connector = ThorConnector(env, session, auto_ban=True)

        for block in range(100000, 500000, 1000):
            pools = await connector.query_pools(block)
            print(f'#{block = }: {len(pools)} pools')


if __name__ == '__main__':
    logging.basicConfig(level=logging.getLevelName(INFO))
    asyncio.run(main())
